'use strict';

var async = require('async');
var Nuget = require('nuget-runner');
var unzip = require('unzip');
var fs = require('fs');
var path = require('path');
var mkdirp = require('mkdirp');
var ejs = require('ejs');
var semver = require('semver');
var _ = require('lodash');
var DownAll = require('down-all');
var rimraf = require('rimraf');
var arti = require('./tools/arti');

var WindowPain = function(runConfig) {
	var self = this;

	var path2root = process.cwd();// This is the path from this file back to the root of the application.

	var appPkg = require(path.join(path2root, 'package.json'));// Read package.json of the main app for which this packaged is installed
	var pkgPkg = require(path.join(__dirname, 'package.json'));// Read package.json of this package
	var config;
	var buildProperties = {};

	function status(step, message) {
		console.log('\n'+'-'.repeat(79)+'\n['+step+'/9] '+message+' ...\n'+'-'.repeat(79)+'\n');
	}

	/**
	 * Initialize configuration.
	 *
	 * @param callback
	 */
	self.initialize = function(callback) {
		status(1, 'Initializing configuration');

		if (!_.has(appPkg, "config.windowPain")) {
			return callback("app package.json does not have config.windowPain");
		}

		var appConfig = _.get(appPkg, 'config.windowPain', {});
		var pkgConfig = _.get(pkgPkg, 'config.windowPain', {});
		config = _.defaultsDeep({}, runConfig, appConfig, pkgConfig);// merge configs from command line, window-pain package.json, app package.json

		var options = _.reduce(JSON.parse(process.env.npm_config_argv).remain, function(args, arg) {
			var pair = arg.split("=");
			args[pair[0]] = pair[1];
			return args;
		}, {});

		//Overwrite version if defined in Jenkins properties
		var j = path.join(path2root, '.build/jenkins.properties');
		if (fs.existsSync(j)) {
			var buildPropertiesString = fs.readFileSync(j).toString();
			var regex = /(.+)=(.+)/g;
			var match;
			while (match = regex.exec(buildPropertiesString)) {
				buildProperties[match[1]] = match[2];
			}
			if (buildProperties.hasOwnProperty('TWO_NUMBERS')) {
				appPkg.version = buildProperties.TWO_NUMBERS+'.0';
			}
		}

		//Override version if passed on command line
		if (options.hasOwnProperty('version')) {
			if (semver.valid(options.version)) {
				appPkg.version = options.version;
			} else {
				options.version = options.version + ".0";
				if (semver.valid(options.version)) {
					var currentPatchLevel = null;//Need to actually go find the current patch level at this major.minor version, or set to 0 if none.
					if (currentPatchLevel != null) {
						appPkg.version = semver.inc(options.version, 'patch');
					} else {
						appPkg.version = options.version;
					}
				}
			}
		}

		//Merge the dowload files from the app package and the window-pain package.
		//The window-pain package acts as a set of default locations for downloaded files.
		//The app package can specify additional URLs for downloaded files.
		//The app locations will be used first. If success then window-pain sources will be ignored.
		if (_.has(appConfig, "download.files")) {
			config.download.files = _.transform(
				_.reduce(
					_.concat(appConfig.download.files, pkgConfig.download.files),//app config needs to be earlier in the concatenated array.
					function(result, file) {
						var key = file.path+"/"+file.name;
						if (!result.hasOwnProperty(key)) {
							result[key] = {
								name: file.name,
								path: file.path,
								urls: []
							};
						}
						if (file.hasOwnProperty("url")) {
							result[key].urls.push(file.url);
						}
						if (file.hasOwnProperty("urls")) {
							_.concat(result[key].urls, file.urls);
						}
						return result;
					},
					{}
				),
				function(result, value) {//This is the function that transforms the reduced result back into an array of files.
					if (value.urls.length == 1) {
						value.url = value.urls[0];
						delete value.urls;
					}
					return result.push(value);
				},
				[]
			);
		}
		return callback(null);
	};

	/**
	 * Remove previously loaded downloads and other build resource folders.
	 *
	 * @param callback
	 */
	self.cleanStep = function(callback) {
		status(2, 'Clean directories used for the build');

		var _clean = function(pathObj, callback) {
			console.log('cleaning ' + pathObj.description + ' in ' + pathObj.path);
			rimraf(pathObj.path, callback);
		};

		async.each(config.clean.paths, _clean, callback);
	};


	/**
	 * Download binary dependencies listed in package.json
	 *
	 * @param callback
	 */
	self.downloadStep = function(callback) {
		status(3, 'Downloading required binaries');

		if (!config.download.overwrite) {
			console.log('package.json option to overwrite binaries disabled...');
		}
		if (!config.download.strictSSL) {
			console.log('package.json option strictSSL disabled...');
		}

		var downAll = DownAll({
			progress : true,
			overwrite : config.download.overwrite,
			dest : config.download.dest,
			strictSSL : config.download.strictSSL
		});

		downAll.on('error', function(e, fileObj) {
			console.log('File download error. File: ', fileObj, ', Error: ', e);
		});

		downAll.load(config.download.files, function(err, msg) {
			if (err) {
				return callback(err);
			}
			//console.log(msg);
			return callback(null);
		});
	};


	/**
	 * Unzip archives
	 *
	 * @param callback
	 */
	self.unzipStep = function(callback) {
		status(4, 'Unzipping');

		async.each(
			config.unzip.files,
			function(file, cb) {
				var srcFile = path.join(config.unzip.src, file.src);
				var dest = path.join(config.download.dest, file.dest);

				// check if it exists and is file
				if (!fs.statSync(srcFile).isFile()) {
					return cb(new Error('ERROR: not a file: ' + srcFile));
				}

				console.log('Unzipping ' + srcFile);

				// Read the archive to stream
				fs.createReadStream(srcFile)
					.on('end', function() {
						console.log('Done ' + srcFile);
						return cb(null);
					})
					.on('error', function(error) {
						console.log('ERROR reading ' + srcFile);
						return cb(error);
					})
					// Pipe resulting files to local path
					.pipe(
						unzip.Extract({
							path : dest
						})
						.on('error', function(error) {
							console.log('ERROR unzipping ' + srcFile + ' to ' + dest);
							return cb(error);
						})
					)
				;
			},
			callback
		);
	};


	/**
	 * Get next patch level from repo
	 *
	 * @param callback
	 */
	self.getNextPatchLevel = function(callback) {
		status(5, 'Getting next patch level');

		var repo = {url: null, type: null};

		if (buildProperties.hasOwnProperty('ARTIFACT_REPOSITORY_URL')) {
			repo.url = buildProperties['ARTIFACT_REPOSITORY_URL'];
		}
		if (buildProperties.hasOwnProperty('ARTIFACT_REPOSITORY_TYPE')) {
			repo.type = buildProperties['ARTIFACT_REPOSITORY_TYPE'];
		}
		arti.getNextPatchLevel(config.nuget.name, appPkg.version, repo, function(err, version) {
			if (err) {
				if (err.code == 'ENOTCONN') {
					console.log('Cannot connect to artifact repository.');
					console.log('Unable to calculate next version.');
					console.log('Will proceed with version 0.0.0.');
					appPkg.version = '0.0.0';//Allows build in disconnected dev environments.
					return callback(null);
				}
				return callback(err);
			}
			appPkg.version = version;
			return callback(null);
		});
	};


	/**
	 * Write package.json to resource directory to run npm install --production
	 *
	 * @param callback
	 */
	self.copyPackageStep = function(callback) {
		status(6, 'Copying package.json');

		var source = path.join(path2root, 'package.json');
		var target = path.join(path2root, 'build/dependencies/', 'package.json');

		mkdirp.sync(path.dirname(target));
		fs.writeFileSync(target, JSON.stringify(appPkg, null, 2), {encoding : 'utf8'});

		return callback(null);
	};


	/**
	 * Run npm install --production
	 *
	 * @param callback
	 */
	self.npmInstallStep = function(callback) {
		status(7, 'Running \'npm install --production\'');

		var spawn = require('child_process').spawn;
		var child = spawn(
			(process.platform === "win32" ? "npm.cmd" : "npm"),
			['install', '--production'],
			{cwd : path.join(path2root, 'build/dependencies/')}
		);
		child.stdout.on('data', function(data) {
//			console.log(data.toString());//Puts useless info in the log.
		});
		child.stderr.on('data', function (data) {
			console.log('stderr: ' + data.toString());
		});
		child.on('message', function (message) {
			console.log(message);
		});
		child.on('exit', function (code) {
			console.log('child process exited with code ' + code.toString());
			return callback();
		});
		child.on('error', function (error) {
			return callback(error);
		});
	};


	/**
	 * Create the resources required for a nuget package
	 *
	 * @param callback
	 */
	self.nugetResourcesStep = function(callback) {
		status(8, 'Building nuget package resources');

		// process each template
		async.each(
			config.resources.templates,
			function(template, cb) {
				//	tasks.push(function(cb) {
				console.log('Building ' + template.target);
				try {
					var srcContent = fs.readFileSync(path.join(__dirname, template.src), {encoding : 'utf8'});
					var targetContent = ejs.render(
						srcContent,
						{	appPkg : appPkg,
							pkgPkg : pkgPkg,
							config : config
							//files:  config.nuget.files
						},
						{}
					);

					//var targetFile = path.join(__dirname, '../..', template.target);
					var targetFile = path.join(path2root, template.target);
					mkdirp.sync(path.dirname(targetFile));
					fs.writeFileSync(targetFile, targetContent, {encoding : 'utf8'});

				} catch (err) {
					console.log('ERROR building templates:', err);
					return cb(err);
				}
				return cb(null);
			},
			callback
		);
	};


	/**
	 * Build NuGet Package with 'nuget pack'
	 *
	 * @param callback
	 */
	self.nugetPackStep = function(callback) {
		status(9, 'Building the NuGet package');
		// add to config.exclude devDependencies from app's package.json
		if (!config.nuget.exclude || !Array.isArray(config.nuget.exclude)) {
			config.nuget.exclude = [];
		}
		mkdirp.sync(config.nuget.nupkgPath);

		Nuget({
			nugetPath : config.nuget.exe, // 'path/to/nuget.exe',
			verbosity: 'normal'
		}).pack({
			spec : config.nuget.nuspec,
			outputDirectory : config.nuget.nupkgPath,
			//basePath: '' // The base path of the files defined in the nuspec file.
			version : appPkg.version,
			exclude : config.nuget.exclude,
			noDefaultExcludes : true,
			noPackageAnalysis : true // Needed to prevent nuget warning about libraries (dlls) outside of the lib folder.
		}).fail(function(error) {
			return callback(error);
		}).done(function(stdout) {
			var srcContent = fs.readFileSync(path.join(__dirname, "tools/deploy.cmd"), {encoding : 'utf8'});
			var targetContent = ejs.render(
				srcContent,
				{config : config},
				{}
			);
			var targetFile = path.join(config.nuget.nupkgPath, "deploy.cmd");
			fs.writeFileSync(targetFile, targetContent, {encoding : 'utf8'});
			return callback(null);
		});
	};


	/**
	 * Last Step. Print out the results
	 *
	 * @param error
	 * @param results
	 */
	self.resultsStep = function(error, results) {
		if ( typeof error === 'string') {
			error = {
				message : error
			};
		}
		if (error) {
			console.log('\n*******************************************************************************');
			console.log('* Build interrupted:\nError: ' + error.message + '\nStack:' + error.stack);
			console.log('*******************************************************************************');
		} else {
			console.log('\n*******************************************************************************');
			console.log('* Build finished.');
			console.log('*******************************************************************************');
		}
	};


	self.build = function () {
		//validatePackageConfiguration();

		async.series(
			[
				self.initialize,
				self.cleanStep,
				self.downloadStep,
				self.unzipStep,
				self.getNextPatchLevel,
				self.copyPackageStep,
				self.npmInstallStep,
				self.nugetResourcesStep,
				self.nugetPackStep
			],
			self.resultsStep
		);
	};
};


module.exports = function(runConfig) {
	return new WindowPain(runConfig);
};
